import re
from datetime import datetime
from enum import StrEnum, auto
from functools import cached_property
from pathlib import Path

from jinja2 import Template

from autosd_demo.cli.exceptions import ValidationError
from autosd_demo.core.build.base import BaseEnv
from autosd_demo.core.osbuild import OSBuildConfig
from autosd_demo.utils import in_autosd_demo


class SignedType(StrEnum):
    TRUE = auto()
    FALSE = auto()
    INVALID = auto()

    @classmethod
    def _missing_(cls, value):
        if isinstance(value, bool):
            return cls(str(value).lower())
        match value:
            case "True" | "yes":
                return cls.TRUE
            case "False" | "no":
                return cls.FALSE
            case _:
                return cls.INVALID


class BuildEnv(BaseEnv):
    LINK_PATTERN = re.compile(r"^[a-z]+://")

    def __init__(self, host, osbuild=False, build_id=None):
        super().__init__()
        self.host = host
        self._set_build_env_config(build_id)
        if osbuild:
            self.update(
                {
                    "osbuild": dict(
                        OSBuildConfig(self.build_files, self.build_directories)
                    )
                }
            )

    @property
    def is_remote(self):
        return self.host is not None

    @cached_property
    def base_dir(self):
        if self.is_remote:
            return (
                self.conf["remote_host"][self.host]["base_dir"]
                if self.conf_exists("remote_host", self.host, "base_dir")
                else "~"
            )

        return self.conf["base_dir"]

    @property
    def build_id(self):
        return self["build"]["build_id"]

    @property
    def hooks(self):
        return self["hooks"]

    @cached_property
    def build_dir(self):
        if self.is_remote:
            return (
                self.conf["remote_host"][self.host]["build_dir"]
                if self.conf_exists("remote_host", self.host, "build_dir")
                else str(Path(self.base_dir) / "build")
            )

        return self.conf["build_dir"]

    @cached_property
    def build_id_path(self):
        return Path(self.build_dir, self.build_id)

    @property
    def build_src_dir(self):
        return self.build_id_path / "src"

    @property
    def registry_auth(self):
        return self.build_id_path / "auth.json"

    @cached_property
    def containers_dir(self):
        return (
            Path(self.build_dir) / "cache" / "containers"
            if self.conf["cache"]["containers_storage"]
            else self.build_id_path / ".containers"
        )

    @property
    def storage_conf(self):
        return self.containers_dir / "storage.conf"

    @property
    def containers_conf(self):
        return self.containers_dir / "containers.conf"

    def cache_dir(self, cached):
        return (
            Path(self.build_dir) / "cache" / "aib"
            if cached
            else self.build_id_path / ".aib"
        )

    @property
    def build_cache_dir(self):
        return self.cache_dir(self.conf["cache"].get("aib_build", False)) / "build"

    @property
    def mpp_cache_dir(self):
        return self.cache_dir(self.conf["cache"].get("aib_mpp", False)) / "mpp"

    @property
    def build_files(self):
        extra_contents_list = self["build"]["src"]["files"]
        extra_contents_list += self["build"]["src"]["templated_files"]
        return sorted(extra_contents_list, key=lambda x: x["dest"])

    @property
    def build_directories(self):
        return self["build"]["src"].get("directories", [])

    @property
    def max_cache_size(self):
        return self.conf["cache"].get("aib_max_size", "2GB")

    def _set_build_env_config(self, build_id):
        self.update({"conf": self.conf, "build": {}, "hooks": {}})
        self["build"]["build_id"] = (
            build_id if build_id else f"{datetime.now():%Y%m%d_%H%M%S}"
        )
        self["build"]["is_remote"] = self.is_remote
        if in_autosd_demo():
            self["build"]["base_dir"] = self.base_dir
            self["build"]["build_dir"] = self.build_dir
            self._get_configured_hooks()
            self._reset_build_src_contents()
            self._get_build_src_extra_files()
            self._process_containers()
            self._process_apps()

    def _get_configured_hooks(self):
        stages = ["setup", "pre_build", "post_build", "clean_build"]
        self["hooks"] = {stage: [] for stage in stages}

        if not self.conf_exists("hooks"):
            return

        for stage in stages:
            hooks = []
            for hook_data in self.conf["hooks"].values():
                hooks.extend(
                    [
                        (
                            hook
                            if Path(hook).is_absolute()
                            # Hooks do not need remote base_dir, so we
                            # always use the conf variable.
                            else str(Path(self.conf["base_dir"]) / hook)
                        )
                        for hook in hook_data.get(stage, [])
                    ]
                )
            self["hooks"][stage] = sorted(hooks, key=lambda path: Path(path).name)

    def reload_src_contents(self):
        self._reset_build_src_contents()
        self._get_build_src_extra_files()
        if self.conf_exists("containers"):
            for container_id, container_data in self.conf["containers"].items():
                if "container_context" in container_data:
                    self._get_build_src_local_containers_files(
                        container_data, container_id
                    )
                if container_data.get("quadlet", True):
                    self._get_build_src_quadlets(container_data, container_id)
            self.reduce_directory_list(self["build"]["src"], "directories")

    def _reset_build_src_contents(self):
        self["build"].setdefault("src", {}).update(
            {
                "directories": [],
                "files": [],
                "templated_files": [],
                "fetch_files": [],
            }
        )

    def _process_apps(self):
        self["build"].setdefault("apps", [])
        if self.conf_exists("apps"):
            for app_id, app_data in self.conf["apps"].items():
                # Init app_spec
                app_spec = app_data.copy()
                app_spec["id"] = app_id
                app_spec.setdefault("info", "Information N/A.")
                app_spec["src"] = {
                    "directories": [],
                    "files": [],
                    "templated_files": [],
                }
                app_spec["build"] = {"directories": [], "templated_files": []}
                app_spec["paths"] = {}

                app_paths = app_spec["paths"]
                app_paths["prefix"] = f"{app_spec['name']}-{app_spec['version']}"
                app_paths["src"] = f"{app_spec['paths']['prefix']}/src"
                app_paths["dist"] = f"{app_spec['paths']['prefix']}/dist"

                # Shortcut for templates
                templates = self.conf["centos_sig"]["autosd_demo"]["templates"]

                # Add build config assets
                app_spec["build"]["directories"].extend(
                    [
                        f"{app_spec['id']}/containers",
                        f"{app_spec['id']}/containers/storage",
                    ]
                )
                app_paths[
                    "containers_conf"
                ] = f"{app_spec['id']}/containers/containers.conf"
                app_paths["storage_conf"] = (
                    f"{app_spec['id']}/" "containers/storage.conf"
                )
                app_spec["build"]["templated_files"].extend(
                    [
                        {
                            "src": templates["containers"]["containers_conf_path"],
                            "dest": app_paths["containers_conf"],
                            "app_id": app_spec["id"],
                        },
                        {
                            "src": templates["containers"]["storage_conf_path"],
                            "dest": app_paths["storage_conf"],
                            "app_id": app_spec["id"],
                        },
                    ]
                )

                # Populate app's containers assets and build config
                layers = set()
                for container_id, container_data in app_data["containers"].items():
                    container_data["id"] = container_id
                    layer = container_data.get("layer", "root")
                    layers.add(layer)

                    if "container_context" in container_data:
                        self._get_build_src_local_containers_files(
                            container_data, container_id, app_spec
                        )

                    if container_data.get("quadlet", True):
                        container_data["quadlet_file"] = (
                            Path(app_spec["paths"]["dist"])
                            / "var"
                            / "apps"
                            / app_spec["name"]
                            / app_spec["version"]
                            / (Path("etc") if layer == "root" else Path("etc") / layer)
                            / "containers"
                            / "systemd"
                            / (
                                f'{container_data.get("quadlet_name", container_id)}'
                                f".container"
                            )
                        ).as_posix()

                        self._get_build_src_quadlets(
                            container_data, container_id, app_spec
                        )

                    dist_prefix = Path(app_paths["dist"])
                    var_prefix = Path("var") if layer == "root" else Path("var") / layer
                    app_prefix = (
                        var_prefix / "apps" / app_spec["name"] / app_spec["version"]
                    )

                    container_data["build"] = {
                        "containers_conf": app_paths["containers_conf"],
                        "storage_conf": app_paths["storage_conf"],
                        "src_path": f"{app_spec['paths']['src']}/{container_id}",
                        "dist_path": (
                            dist_prefix / app_prefix / "containers" / "storage"
                        ).as_posix(),
                    }

                # Populate dist assets
                for layer in layers:
                    dist_prefix = Path(app_paths["dist"])
                    var_prefix = Path("var") if layer == "root" else Path("var") / layer
                    app_prefix = (
                        var_prefix / "apps" / app_spec["name"] / app_spec["version"]
                    )
                    app_spec["src"]["directories"].extend(
                        [
                            (dist_prefix / "var" / "images").as_posix(),
                            (dist_prefix / app_prefix / "containers").as_posix(),
                            (
                                dist_prefix / app_prefix / "containers" / "storage"
                            ).as_posix(),
                        ]
                    )

                # Convert containers field into a list for building
                app_spec["containers"] = list(app_spec["containers"].values())

                # Clean up app_spec
                app_spec["src"]["directories"] = sorted(
                    set(app_spec["src"]["directories"])
                )
                app_spec["src"]["files"].sort(key=lambda x: x["dest"])
                app_spec["src"]["templated_files"].sort(key=lambda x: x["dest"])
                app_spec["build"]["directories"] = sorted(
                    set(app_spec["build"]["directories"])
                )
                app_spec["build"]["templated_files"].sort(key=lambda x: x["dest"])

                # Add app_spec to the build.apps
                self["build"]["apps"].append(app_spec)

    def _process_containers(self):
        self["build"].setdefault("containers", [])
        if self.conf_exists("containers"):
            for container_id, container_data in self.conf["containers"].items():
                container_spec = container_data.copy()
                container_spec["id"] = container_id

                if "container_context" in container_data:
                    self._get_build_src_local_containers_files(
                        container_data, container_id
                    )

                if container_data.get("quadlet", True):
                    self._get_build_src_quadlets(container_data, container_id)

                self["build"]["containers"].append(container_spec)
            self.reduce_directory_list(self["build"]["src"], "directories")

    def _get_build_src_local_containers_files(
        self, container_data, container_id, app_spec=None
    ):
        build_src = app_spec["src"] if app_spec else self["build"].setdefault("src", {})
        context_path = Path(container_data["container_context"])
        if not context_path.is_absolute():
            context_path = Path(self.conf["base_dir"]) / context_path

        list_files = [f for f in context_path.glob("**/*") if f.is_file()]

        for file in list_files:
            path_template = Template(
                file.as_posix(),
                variable_start_string="{",
                variable_end_string="}",
            )
            src = Path(path_template.render(conf=self.conf))
            if app_spec is None:
                dest = Path("files") / self.conf["name"]
                dest = dest / "containers" / container_id
                dest = dest / src.relative_to(context_path)
            else:
                dest = Path(app_spec["paths"]["src"])
                dest = dest / container_id / src.relative_to(context_path)

            if file.suffix == ".j2":
                build_src.setdefault("templated_files", []).append(
                    {
                        "src": file.as_posix(),
                        "dest": dest.with_suffix("").as_posix(),
                    }
                )
            else:
                build_src.setdefault("files", []).append(
                    {"src": file.as_posix(), "dest": dest.as_posix()}
                )
            build_src.setdefault("directories", []).append(dest.parent.as_posix())
        build_src["directories"] = sorted(set(build_src["directories"]))

    def _get_build_src_extra_files(self):
        if self.conf_exists("extra_files"):
            for ef_id, ef_data in self.conf["extra_files"].items():
                touch = ef_data.get("touch", False)
                delete = ef_data.get("delete", False)
                if delete and touch:
                    raise ValidationError(
                        "extra_files",
                        ef_id,
                        "'touch' and 'delete' cannot be enabled simultaneously",
                    )
                if not (delete or touch):
                    self.conf_required("extra_files", ef_id, "src")
                self.conf_required("extra_files", ef_id, "dest")

                dest_path = (
                    Path(ef_data["dest"][1:])
                    if Path(ef_data["dest"]).is_absolute()
                    else Path(ef_data["dest"])
                )
                if touch:
                    self._get_build_src_extra_files_touch(ef_data, dest_path)
                elif delete:
                    self._get_build_src_extra_files_delete(ef_data, dest_path)
                elif self.LINK_PATTERN.match(ef_data["src"]):
                    self._get_build_src_extra_files_link(ef_data, dest_path)
                else:
                    self._get_build_src_extra_files_regular(ef_data, dest_path)

    def _get_build_src_extra_files_touch(self, ef_data, dest_path):
        build_src = self["build"].setdefault("src", {})
        layer = ef_data.get("layer", "root")

        dest = Path("files") / self.conf["name"]
        dest = dest / "extra_files" / f"{layer}_fs"
        dest = dest / dest_path

        if ef_data["dest"].endswith("/"):
            build_src["directories"].append(dest.as_posix())
        else:
            build_src["files"].append({"dest": dest.as_posix()})
            build_src["directories"].append(dest.parent.as_posix())

    def _get_build_src_extra_files_delete(self, ef_data, dest_path):
        build_src = self["build"].setdefault("src", {})
        layer = ef_data.get("layer", "root")

        dest = Path("files") / self.conf["name"]
        dest = dest / "delete" / f"{layer}_fs"
        dest = dest / dest_path

        build_src["files"].append({"dest": dest.as_posix()})

    def _get_build_src_extra_files_link(self, ef_data, dest_path):
        build_src = self["build"].setdefault("src", {})
        layer = ef_data.get("layer", "root")

        url = ef_data["src"]
        checksum = ef_data["checksum"] if "checksum" in ef_data else ""
        file_name = Path(url.replace("://", "_").replace("/", "_"))
        src = Path(self.conf["base_dir"]) / "build" / "cache" / "downloads" / file_name
        dest = (
            Path("files")
            / self.conf["name"]
            / "extra_files"
            / f"{layer}_fs"
            / dest_path
        )
        if ef_data["dest"].endswith("/"):
            dest_name = url.rpartition("/")[2]
            dest = dest / dest_name

        build_src["fetch_files"].append(
            {"url": url, "checksum": checksum, "file": file_name.as_posix()}
        )
        build_src["files"].append({"src": src.as_posix(), "dest": dest.as_posix()})
        build_src["directories"].append(dest.parent.as_posix())

    def _get_build_src_extra_files_regular(self, ef_data, dest_path):
        build_src = self["build"].setdefault("src", {})
        layer = ef_data.get("layer", "root")
        src_path = Path(ef_data["src"])

        if not src_path.is_absolute():
            src_path = Path(self.conf["base_dir"]) / src_path

        if src_path.is_file():
            list_files = [src_path]
        elif src_path.is_dir():
            list_files = [f for f in src_path.glob("**/*") if f.is_file()]
        else:
            return

        for file in list_files:
            path_template = Template(
                file.as_posix(),
                variable_start_string="{",
                variable_end_string="}",
            )
            src = Path(path_template.render(conf=self.conf))
            dest = Path("files") / self.conf["name"]
            dest = dest / "extra_files" / f"{layer}_fs"
            dest = dest / dest_path

            if len(list_files) == 1 and ef_data["dest"].endswith("/"):
                dest = dest / src.name
            else:
                dest = dest / src.relative_to(src_path)

            if file.suffix == ".j2":
                if (
                    len(list_files) > 1
                    or not src_path.is_file()
                    or ef_data["dest"].endswith("/")
                ):
                    dest = dest.with_suffix("")
                build_src["templated_files"].append(
                    {
                        "src": file.as_posix(),
                        "dest": dest.as_posix(),
                    }
                )
            else:
                build_src["files"].append(
                    {"src": file.as_posix(), "dest": dest.as_posix()}
                )
            build_src["directories"].append(dest.parent.as_posix())

    def _get_build_src_quadlets(self, container_data, container_id, app_spec=None):
        build_src = app_spec["src"] if app_spec else self["build"].setdefault("src", {})
        layer = container_data.get("layer", "root")
        if app_spec is None:
            quadlet_dest = (
                Path("files") / self.conf["name"] / "extra_files" / f"{layer}_fs"
            )
            quadlet_dest = (
                quadlet_dest / f"etc/containers/systemd/{container_id}.container"
            )
        else:
            quadlet_dest = Path(container_data["quadlet_file"])

        centos_sig_templates = self.conf["centos_sig"]["autosd_demo"]["templates"]
        quadlet_template = {
            "src": centos_sig_templates["containers"]["quadlet"],
            "dest": quadlet_dest.as_posix(),
            "container_id": container_id,
            "sign": SignedType(container_data.get("sign", "true")).value,
        }
        if app_spec:
            quadlet_template["app_id"] = app_spec["id"]
            # image_storage_path doesn't have visibility of the layers because
            # the quadlet file runs inside a layer, so always it'll check for
            # the image at /var/apps.
            quadlet_template["image_storage_path"] = (
                Path("/var")
                / "apps"
                / app_spec["name"]
                / app_spec["version"]
                / "containers"
                / "storage"
            ).as_posix()

        build_src.setdefault("templated_files", []).append(quadlet_template)
        build_src.setdefault("directories", []).append(quadlet_dest.parent.as_posix())
